from collections import deque
from typing import Deque, Dict, List, Optional, Set

from scaled.scheduler.worker_manager.allocators.mixins import TaskAllocator


class QueuedAllocator(TaskAllocator):
    def __init__(self, max_tasks_per_worker: int):
        self._max_tasks_per_worker = max_tasks_per_worker
        self._workers_to_task_ids: Dict[bytes, Set[bytes]] = dict()
        self._capacity: Deque[bytes] = deque()
        self._task_to_worker = {}

    def add_worker(self, worker: bytes) -> bool:
        if worker in self._workers_to_task_ids:
            return False

        self._workers_to_task_ids[worker] = set()
        self._capacity.extend([worker] * self._max_tasks_per_worker)
        return True

    def remove_worker(self, worker: bytes) -> List[bytes]:
        if worker not in self._workers_to_task_ids:
            return []

        new_capacity = deque()
        new_capacity.extend([w for w in self._capacity if w != worker])
        self._capacity = new_capacity

        tasks = self._workers_to_task_ids.pop(worker)
        for task in tasks:
            self._task_to_worker.pop(task)
        return list(tasks)

    def assign_task(self, task_id: bytes) -> Optional[bytes]:
        if not self._capacity:
            return None

        if task_id in self._task_to_worker:
            return self._task_to_worker[task_id]

        worker = self._capacity.popleft()
        self._workers_to_task_ids[worker].add(task_id)
        self._task_to_worker[task_id] = worker
        return worker

    def remove_task(self, task_id: bytes) -> Optional[bytes]:
        if task_id not in self._task_to_worker:
            return None

        worker = self._task_to_worker.pop(task_id)
        self._workers_to_task_ids[worker].remove(task_id)
        self._capacity.append(worker)
        return worker

    def get_assigned_worker(self, task_id: bytes) -> Optional[bytes]:
        if task_id not in self._task_to_worker:
            return None

        return self._task_to_worker[task_id]

    def statistics(self) -> Dict:
        return {
            "type": "queued",
            "worker_to_tasks": {worker.decode(): len(tasks) for worker, tasks in self._workers_to_task_ids.items()},
        }



import logging
import os
import socket
from collections import defaultdict
from typing import Awaitable, Callable, List, Literal

import zmq.asyncio

from scaled.io.config import POLLING_TIME_MILLISECONDS
from scaled.utility.zmq_config import ZMQConfig
from scaled.protocol.python.message import MessageType, MessageVariant, PROTOCOL


class AsyncConnector:
    def __init__(
        self,
        prefix: str,
        context: zmq.asyncio.Context,
        socket_type: int,
        address: ZMQConfig,
        bind_or_connect: Literal["bind", "connect"],
        callback: Callable[[MessageType, MessageVariant], Awaitable[None]],
    ):
        self._prefix = prefix
        self._address = address

        self._context = context
        self._socket = self._context.socket(socket_type)
        self._identity: bytes = f"{self._prefix}|{socket.gethostname()}|{os.getpid()}".encode()
        self.__set_socket_options()

        if bind_or_connect == "bind":
            self._socket.bind(self._address.to_address())
        elif bind_or_connect == "connect":
            self._socket.connect(self._address.to_address())
        else:
            raise TypeError(f"bind_or_connect has to be 'bind' or 'connect'")

        self._callback: Callable[[MessageType, MessageVariant], Awaitable[None]] = callback

        self._statistics = {"received": defaultdict(lambda: 0), "sent": defaultdict(lambda: 0)}

    def __del__(self):
        self._socket.close()

    @property
    def identity(self) -> bytes:
        return self._identity

    async def routine(self):
        count = await self._socket.poll(POLLING_TIME_MILLISECONDS)
        if not count:
            return

        for _ in range(count):
            frames = await self._socket.recv_multipart()
            if not self.__is_valid_message(frames):
                continue

            message_type_bytes, *payload = frames
            message_type = MessageType(message_type_bytes)
            message = PROTOCOL[message_type_bytes].deserialize(payload)

            self.__count_one("received", message_type)
            await self._callback(message_type, message)

    async def send(self, message_type: MessageType, data: MessageVariant):
        self.__count_one("sent", message_type)
        await self._socket.send_multipart([message_type.value, *data.serialize()])

    async def statistics(self):
        return self._statistics

    def __set_socket_options(self):
        self._socket.setsockopt(zmq.IDENTITY, self._identity)
        self._socket.setsockopt(zmq.SNDHWM, 0)
        self._socket.setsockopt(zmq.RCVHWM, 0)

    def __count_one(self, count_type: Literal["sent", "received"], message_type: MessageType):
        self._statistics[count_type][message_type.name] += 1

    def __is_valid_message(self, frames: List[bytes]) -> bool:
        if len(frames) < 2:
            logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
            return False

        if frames[0] not in {member.value for member in MessageType}:
            logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
            return False

        return True

    def __get_prefix(self):
        return f"{self.__class__.__name__}[{self._identity.decode()}]:"



import asyncio
import logging
import multiprocessing
import signal
import threading
import time
from queue import Queue
from collections import defaultdict
from typing import Callable, Dict, List, Optional

import zmq
import zmq.asyncio

from scaled.io.sync_connector import SyncConnector
from scaled.protocol.python.serializer.mixins import FunctionSerializerType
from scaled.utility.event_loop import register_event_loop
from scaled.utility.zmq_config import ZMQConfig, ZMQType
from scaled.protocol.python.message import (
    FunctionRequestType,
    FunctionRequest,
    FunctionResponse,
    FunctionResponseType,
    MessageType,
    MessageVariant,
    Task,
    TaskCancel,
    TaskResult,
    TaskStatus,
)
from scaled.worker.async_agent import AsyncAgent


class Agent(threading.Thread):
    def __init__(
        self,
        stop_event: threading.Event,
        context: zmq.Context,
        address: ZMQConfig,
        address_internal: ZMQConfig,
        heartbeat_interval_seconds: int,
        event_loop: str,
    ):
        threading.Thread.__init__(self)

        self._stop_event: threading.Event = stop_event
        self._address: ZMQConfig = address
        self._address_internal: ZMQConfig = address_internal

        self._heartbeat_interval_seconds: int = heartbeat_interval_seconds

        self._event_loop = event_loop

        self._agent = AsyncAgent(
            stop_event=self._stop_event,
            context=context,
            address=self._address,
            address_internal=self._address_internal,
            heartbeat_interval_seconds=self._heartbeat_interval_seconds,
        )

    def run(self) -> None:
        register_event_loop(self._event_loop)
        asyncio.run(self._agent.loop())


class Worker(multiprocessing.get_context("spawn").Process):
    def __init__(
        self,
        address: ZMQConfig,
        stop_event: multiprocessing.Event,
        heartbeat_interval_seconds: int,
        event_loop: str,
        serializer: FunctionSerializerType,
    ):
        multiprocessing.Process.__init__(self, name="Worker")

        self._address = address
        self._heartbeat_interval_seconds = heartbeat_interval_seconds
        self._event_loop = event_loop
        self._stop_event = stop_event
        self._serializer = serializer

        self._task_queue: Optional[Queue] = None
        self._agent_connector: Optional[SyncConnector] = None
        self._agent: Optional[Agent] = None
        self._ready_event = multiprocessing.get_context("spawn").Event()

        self._cached_functions: Dict[bytes, Callable] = {}
        self._pending_function_to_tasks: Dict[bytes, List[Task]] = defaultdict(list)

    def wait_till_ready(self):
        while not self._ready_event.is_set():
            continue

    def run(self) -> None:
        self.__initialize()
        self.__run_forever()

    def shutdown(self, *args):
        assert args is not None
        self._agent_connector.join()
        self._agent.join()

    def get_prefix(self):
        return f"{self.__class__.__name__}[{self._agent_connector.identity.decode()}]:"

    def __initialize(self):
        self.__register_signal()

        self._task_queue = Queue()

        context = zmq.Context.instance()
        internal_channel = ZMQConfig(type=ZMQType.inproc, host="memory")
        self._agent_connector = SyncConnector(
            stop_event=self._stop_event,
            prefix="A",
            context=context,
            socket_type=zmq.PAIR,
            bind_or_connect="connect",
            address=internal_channel,
            callback=self.__on_connector_receive,
        )
        self._agent = Agent(
            stop_event=self._stop_event,
            context=zmq.asyncio.Context.shadow(context.underlying),
            address=self._address,
            address_internal=internal_channel,
            heartbeat_interval_seconds=self._heartbeat_interval_seconds,
            event_loop=self._event_loop,
        )
        self._agent.start()

        # worker is ready
        self._ready_event.set()

    def __run_forever(self):
        while not self._stop_event.is_set():
            self.__process_queued_tasks()

    def __register_signal(self):
        signal.signal(signal.SIGINT, self.shutdown)
        signal.signal(signal.SIGTERM, self.shutdown)

    def __process_queued_tasks(self):
        if self._task_queue.empty():
            time.sleep(0.1)
            return

        task = self._task_queue.get()
        if task.function_id not in self._cached_functions:
            self._agent_connector.send(
                MessageType.FunctionRequest, FunctionRequest(FunctionRequestType.Request, task.function_id, b"")
            )
            self._pending_function_to_tasks[task.function_id].append(task)
            return

        try:
            function = self._cached_functions[task.function_id]
            if function is None:
                raise ValueError(f"{self.get_prefix()}cannot get function for {task=}")

            args = self._serializer.deserialize_arguments(task.function_args)
            result = self._serializer.serialize_result(function(*args))
            self._agent_connector.send(MessageType.TaskResult, TaskResult(task.task_id, TaskStatus.Success, result))
        except Exception as e:
            logging.exception(f"{self.get_prefix()} error when processing {task=}:")
            self._agent_connector.send(
                MessageType.TaskResult, TaskResult(task.task_id, TaskStatus.Failed, str(e).encode())
            )

    def __on_connector_receive(self, message_type: MessageType, message: MessageVariant):
        if message_type == MessageType.FunctionResponse:
            self.__connector_process_function_response(message)
            return

        if message_type == MessageType.Task:
            self.__connector_process_task(message)
            return

        if message_type == MessageType.TaskCancel:
            self.__connector_process_task_cancel(message)
            return

        logging.error(f"{self.get_prefix()} unsupported {message_type=} {message=}")

    def __connector_process_function_response(self, function_response: FunctionResponse):
        if function_response.status == FunctionResponseType.NotExists:
            return

        assert function_response.status == FunctionResponseType.OK
        self._cached_functions[function_response.function_id] = self._serializer.deserialize_function(
            function_response.content
        )

        if function_response.function_id not in self._pending_function_to_tasks:
            return

        for task in self._pending_function_to_tasks.pop(function_response.function_id):
            self.__connector_process_task(task)

    def __connector_process_task(self, task: Task):
        self._task_queue.put(task)

    def __connector_process_task_cancel(self, task_cancel: TaskCancel):
        # TODO: implement this
        raise NotImplementedError()



import copy
import logging
import os
import socket
import threading
from queue import Queue
from collections import defaultdict
from typing import Callable, Literal

import zmq

from scaled.io.config import POLLING_TIME_MILLISECONDS
from scaled.utility.zmq_config import ZMQConfig
from scaled.protocol.python.message import MessageType, MessageVariant, PROTOCOL


class SyncConnector(threading.Thread):
    def __init__(
        self,
        stop_event: threading.Event,
        prefix: str,
        context: zmq.Context,
        socket_type: int,
        bind_or_connect: Literal["bind", "connect"],
        address: ZMQConfig,
        callback: Callable[[MessageType, MessageVariant], None],
    ):
        threading.Thread.__init__(self)

        self._prefix = prefix
        self._address = address

        self._context = context
        self._socket = self._context.socket(socket_type)
        self._identity: bytes = f"{self._prefix}|{socket.gethostname()}|{os.getpid()}".encode()
        self.__set_socket_options()

        if bind_or_connect == "bind":
            self._socket.bind(self._address.to_address())
        elif bind_or_connect == "connect":
            self._socket.connect(self._address.to_address())
        else:
            raise TypeError(f"bind_or_connect has to be 'bind' or 'connect'")

        self._callback = callback
        self._stop_event = stop_event

        self._send_queue = Queue()

        self._statistics_mutex = threading.Lock()
        self._statistics = {"received": defaultdict(lambda: 0), "sent": defaultdict(lambda: 0)}

        self.start()

    def __del__(self):
        self._socket.close()

    @property
    def identity(self) -> bytes:
        return self._identity

    def run(self) -> None:
        while not self._stop_event.is_set():
            self.__routine_send()
            self.__routine_receive()

    def send(self, message_type: MessageType, message: MessageVariant):
        self._send_queue.put((message_type, message))

    def monitor(self):
        with self._statistics_mutex:
            return copy.copy(self._statistics)

    def __set_socket_options(self):
        self._socket.setsockopt(zmq.IDENTITY, self._identity)
        self._socket.setsockopt(zmq.SNDHWM, 0)
        self._socket.setsockopt(zmq.RCVHWM, 0)

    def __routine_send(self):
        while not self._send_queue.empty():
            message_type, message = self._send_queue.get()
            self._socket.send_multipart([message_type.value, *message.serialize()])

    def __routine_receive(self):
        count = self._socket.poll(POLLING_TIME_MILLISECONDS)
        if not count:
            return

        for _ in range(count):
            frames = self._socket.recv_multipart()

            if len(frames) < 2:
                logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
                return

            if frames[0] not in {member.value for member in MessageType}:
                logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
                return

            message_type_bytes, *payload = frames
            message_type = MessageType(message_type_bytes)
            message = PROTOCOL[message_type_bytes].deserialize(payload)

            self.__count_one("received", message_type)
            self._callback(message_type, message)

    def __count_one(self, count_type: Literal["sent", "received"], message_type: MessageType):
        with self._statistics_mutex:
            self._statistics[count_type][message_type.name] += 1

    def __get_prefix(self):
        return f"{self.__class__.__name__}[{self._identity.decode()}]:"



import abc
from typing import Dict, List, Optional


class TaskAllocator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def add_worker(self, worker: bytes) -> bool:
        """add worker to worker collection"""
        raise NotImplementedError()

    @abc.abstractmethod
    def remove_worker(self, worker: bytes) -> List[bytes]:
        """remove worker to worker collection, and return list of task_ids of removed worker"""
        raise NotImplementedError()

    @abc.abstractmethod
    def assign_task(self, task_id: bytes) -> Optional[bytes]:
        """assign task_id in allocator, return None means no available worker, otherwise will return worker been
        assigned to"""
        raise NotImplementedError()

    @abc.abstractmethod
    def remove_task(self, task_id: bytes) -> Optional[bytes]:
        """remove task in allocator, return None means not found any worker, otherwise will return worker associate
        with the removed task_id"""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_assigned_worker(self, task_id: bytes) -> Optional[bytes]:
        """get worker that been assigned to this task_id, return None means cannot find the worker assigned to this
        task id"""
        raise NotImplementedError()

    @abc.abstractmethod
    def statistics(self) -> Dict:
        raise NotImplementedError()



import pickle
from typing import Any, Callable, Tuple

import cloudpickle

from scaled.protocol.python.serializer.mixins import Serializer


class DefaultSerializer(Serializer):
    @staticmethod
    def serialize_function(fn: Callable) -> bytes:
        return cloudpickle.dumps(fn)

    @staticmethod
    def deserialize_function(payload: bytes) -> Callable:
        return cloudpickle.loads(payload)

    @staticmethod
    def serialize_arguments(args: Tuple[Any, ...]) -> bytes:
        return pickle.dumps(args, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def deserialize_arguments(payload: bytes) -> Tuple[Any, ...]:
        return pickle.loads(payload)

    @staticmethod
    def serialize_result(result: Any) -> bytes:
        return pickle.dumps(result, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def deserialize_result(payload: bytes) -> Any:
        return pickle.loads(payload)



import random

from scaled.client.client import Client
from scaled.cluster.combo import SchedulerClusterCombo
from scaled.utility.zmq_config import ZMQConfig, ZMQType
from scaled.utility.logging.scoped_logger import ScopedLogger
from scaled.utility.logging.utility import setup_logger


def sleep_print(sec: int):
    return sec * 1


def main():
    setup_logger()
    config = ZMQConfig(type=ZMQType.tcp, host="127.0.0.1", port=2345)

    cluster = SchedulerClusterCombo(address=config, n_workers=10, event_loop="uvloop")
    client = Client(config=config)

    tasks = [random.randint(0, 100) for _ in range(100000)]

    with ScopedLogger(f"submit {len(tasks)} tasks"):
        futures = [client.submit(sleep_print, i) for i in tasks]

    with ScopedLogger(f"gather {len(futures)} results"):
        results = [future.result() for future in futures]

    assert results == tasks

    cluster.shutdown()
    client.disconnect()


if __name__ == "__main__":
    main()



import asyncio
import multiprocessing
import threading
from typing import Literal, Optional

from scaled.scheduler.worker_manager.vanilla import AllocatorType
from scaled.utility.event_loop import register_event_loop
from scaled.utility.zmq_config import ZMQConfig
from scaled.scheduler.main import Scheduler
from scaled.utility.logging.utility import setup_logger


class SchedulerProcess(multiprocessing.get_context("spawn").Process):
    def __init__(
        self,
        address: ZMQConfig,
        stop_event: threading.Event,
        allocator_type: AllocatorType,
        worker_timeout_seconds: int,
        function_timeout_seconds: int,
        event_loop: Literal["builtin", "uvloop"] = "builtin",
    ):
        multiprocessing.Process.__init__(self, name="Scheduler")
        self._address = address
        self._stop_event = stop_event
        self._allocator_type = allocator_type
        self._worker_timeout_seconds = worker_timeout_seconds
        self._function_timeout_seconds = function_timeout_seconds
        self._scheduler: Optional[Scheduler] = None
        self._event_loop = event_loop

    def run(self) -> None:
        # scheduler have its own single process
        setup_logger()
        self._scheduler = Scheduler(
            address=self._address,
            stop_event=self._stop_event,
            allocator_type=self._allocator_type,
            worker_timeout_seconds=self._worker_timeout_seconds,
            function_timeout_seconds=self._function_timeout_seconds,
        )

        register_event_loop(self._event_loop)
        asyncio.run(self._scheduler.loop())



import abc
import enum
import pickle
import struct
from typing import Dict, List, Tuple, TypeVar

import attrs


class MessageType(enum.Enum):
    Task = b"TK"
    TaskEcho = b"TE"
    TaskCancel = b"TC"
    TaskCancelEcho = b"TX"
    TaskResult = b"TR"

    Heartbeat = b"HB"

    FunctionRequest = b"FR"
    FunctionResponse = b"FA"

    MonitorRequest = b"MR"
    MonitorResponse = b"MS"

    @staticmethod
    def allowed_values():
        return {member.value for member in MessageType}


class TaskStatus(enum.Enum):
    Success = b"S"
    Failed = b"F"
    Canceled = b"C"


class TaskEchoStatus(enum.Enum):
    SubmitOK = b"SK"
    CancelOK = b"CK"
    Duplicated = b"DC"
    FunctionNotExists = b"FN"


class FunctionRequestType(enum.Enum):
    Check = b"C"
    Add = b"A"
    Request = b"R"
    Delete = b"D"


class FunctionResponseType(enum.Enum):
    OK = b"OK"
    NotExists = b"NE"
    StillHaveTask = b"HT"
    Duplicated = b"DC"


class _Message(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def serialize(self) -> Tuple[bytes, ...]:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def deserialize(data: List[bytes]):
        raise NotImplementedError()


MessageVariant = TypeVar("MessageVariant", bound=_Message)


@attrs.define
class Task(_Message):
    task_id: bytes
    function_id: bytes
    function_args: bytes

    def serialize(self) -> Tuple[bytes, bytes, bytes]:
        return self.task_id, self.function_id, self.function_args

    @staticmethod
    def deserialize(data: List[bytes]):
        return Task(data[0], data[1], data[2])


@attrs.define
class TaskEcho(_Message):
    task_id: bytes
    status: TaskEchoStatus

    def serialize(self) -> Tuple[bytes, ...]:
        return self.task_id, self.status.value

    @staticmethod
    def deserialize(data: List[bytes]):
        return TaskEcho(data[0], TaskEchoStatus(data[1]))


@attrs.define
class TaskCancel(_Message):
    task_id: bytes

    def serialize(self) -> Tuple[bytes, ...]:
        return self.task_id,

    @staticmethod
    def deserialize(data: List[bytes]):
        return TaskCancel(data[0])


@attrs.define
class TaskCancelEcho(_Message):
    task_id: bytes
    status: TaskEchoStatus

    def serialize(self) -> Tuple[bytes, ...]:
        return self.task_id, self.status.value

    @staticmethod
    def deserialize(data: List[bytes]):
        return TaskCancelEcho(data[0], TaskEchoStatus(data[1]))


@attrs.define
class TaskResult(_Message):
    task_id: bytes
    status: TaskStatus
    result: bytes

    def serialize(self) -> Tuple[bytes, ...]:
        return self.task_id, self.status.value, self.result

    @staticmethod
    def deserialize(data: List[bytes]):
        return TaskResult(data[0], TaskStatus(data[1]), data[2])


@attrs.define
class Heartbeat(_Message):
    cpu_usage: float
    rss_size: int

    def serialize(self) -> Tuple[bytes, ...]:
        return struct.pack("fI", self.cpu_usage, self.rss_size),

    @staticmethod
    def deserialize(data: List[bytes]):
        return Heartbeat(*struct.unpack("fI", data[0]))


@attrs.define
class MonitorRequest(_Message):
    def serialize(self) -> Tuple[bytes, ...]:
        return b"",

    @staticmethod
    def deserialize(data: List[bytes]):
        return MonitorRequest()


@attrs.define
class MonitorResponse(_Message):
    data: Dict

    def serialize(self) -> Tuple[bytes, ...]:
        return pickle.dumps(self.data),

    @staticmethod
    def deserialize(data: List[bytes]):
        return MonitorResponse(pickle.loads(data[0]))


@attrs.define
class FunctionRequest(_Message):
    type: FunctionRequestType
    function_id: bytes
    content: bytes

    def serialize(self) -> Tuple[bytes, ...]:
        return self.type.value, self.function_id, self.content

    @staticmethod
    def deserialize(data: List[bytes]):
        return FunctionRequest(FunctionRequestType(data[0]), data[1], data[2])


@attrs.define
class FunctionResponse(_Message):
    status: FunctionResponseType
    function_id: bytes
    content: bytes

    def serialize(self) -> Tuple[bytes, ...]:
        return self.status.value, self.function_id, self.content

    @staticmethod
    def deserialize(data: List[bytes]):
        return FunctionResponse(FunctionResponseType(data[0]), data[1], data[2])


PROTOCOL = {
    MessageType.Heartbeat.value: Heartbeat,
    MessageType.Task.value: Task,
    MessageType.TaskEcho.value: TaskEcho,
    MessageType.TaskCancel.value: TaskCancel,
    MessageType.TaskCancelEcho.value: TaskCancelEcho,
    MessageType.TaskResult.value: TaskResult,
    MessageType.MonitorRequest.value: MonitorRequest,
    MessageType.MonitorResponse.value: MonitorResponse,
    MessageType.FunctionRequest.value: FunctionRequest,
    MessageType.FunctionResponse.value: FunctionResponse,
}



import abc
from typing import Dict

from scaled.protocol.python.message import FunctionRequest, Heartbeat, Task, TaskResult


class Looper(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def routine(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def statistics(self) -> Dict:
        raise NotImplementedError()


class FunctionManager(Looper):
    @abc.abstractmethod
    async def on_function(self, source: bytes, request: FunctionRequest):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_use_function(self, task_id: bytes, function_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_done_function(self, task_id: bytes, function_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def has_function(self, function_id: bytes) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    async def routine(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def statistics(self) -> Dict:
        raise NotImplementedError()


class ClientManager(Looper):
    @abc.abstractmethod
    async def on_task_new(self, client: bytes, task: Task):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_cancel(self, client: bytes, task_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_done(self, result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    async def routine(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def statistics(self) -> Dict:
        raise NotImplementedError()


class TaskManager(Looper):
    @abc.abstractmethod
    async def on_task_new(self, client: bytes, task: Task):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_reroute(self, task_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_cancel(self, client: bytes, task_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_done(self, result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    async def routine(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def statistics(self) -> Dict:
        raise NotImplementedError()


class WorkerManager(Looper):
    @abc.abstractmethod
    async def assign_task_to_worker(self, task: Task) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_cancel(self, task_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_done(self, task_result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_heartbeat(self, worker: bytes, info: Heartbeat):
        raise NotImplementedError()

    @abc.abstractmethod
    async def routine(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def statistics(self) -> Dict:
        raise NotImplementedError()



from typing import Dict, Generator, List, Optional, Set
import logging

from scaled.scheduler.worker_manager.allocators.mixins import TaskAllocator


class OneToOneAllocator(TaskAllocator):
    def __init__(self):
        self._workers = WorkerCollection()
        self._task_to_worker = {}

    def add_worker(self, worker: bytes) -> bool:
        if self._workers.has_worker(worker):
            return False

        self._workers[worker] = None
        return True

    def remove_worker(self, worker: bytes) -> List[bytes]:
        task_id = self._workers.pop(worker)
        if task_id is None:
            return []

        return [task_id]

    def assign_task(self, task_id: bytes) -> Optional[bytes]:
        if self._workers.full():
            return None

        worker = self._workers.get_one_unused_worker()
        self._workers[worker] = task_id
        self._task_to_worker[task_id] = worker
        return worker

    def get_assigned_worker(self, task_id: bytes) -> Optional[bytes]:
        return self._task_to_worker.get(task_id, None)

    def remove_task(self, task_id: bytes) -> Optional[bytes]:
        if task_id not in self._task_to_worker:
            logging.error(f"received {task_id=} not in workers")
            return None

        worker = self._task_to_worker.pop(task_id)
        self._workers[worker] = None
        return worker

    def statistics(self) -> Dict:
        return {
            "type": "one_to_one",
            "unused_workers": [worker.decode() for worker in self._workers.get_unused_workers()],
            "used_workers": {k.decode(): v.decode() for k, v in self._workers.get_used_workers().items()},
        }


class WorkerCollection:
    def __init__(self):
        self._used_workers: Dict[bytes, bytes] = {}
        self._unused_workers: Set[bytes] = set()

    def __setitem__(self, worker: bytes, task: Optional[bytes]):
        assert isinstance(worker, bytes)
        if task is not None:
            if worker in self._used_workers:
                raise ValueError(f"assign {task=} to {worker=} while it still processing {self._used_workers[worker]}")

            self._unused_workers.remove(worker)
            self._used_workers[worker] = task
        else:
            if worker in self._used_workers:
                self._used_workers.pop(worker)

            self._unused_workers.add(worker)

    def __getitem__(self, worker: bytes) -> Optional[bytes]:
        if not isinstance(worker, bytes):
            raise TypeError(f"Expect worker to be type bytes, got: type={type(worker)} {worker=}")

        return self._used_workers.get(worker, None)

    def keys(self) -> Generator[bytes, None, None]:
        yield from self._unused_workers
        yield from self._used_workers.keys()

    def pop(self, worker: bytes) -> Optional[bytes]:
        if worker not in self._used_workers and worker not in self._unused_workers:
            raise KeyError(f"worker not exists: {worker}")

        if worker in self._used_workers:
            return self._used_workers.pop(worker)

        self._unused_workers.remove(worker)
        return None

    def size(self) -> int:
        return len(self._used_workers) + len(self._unused_workers)

    def capacity(self) -> int:
        return len(self._unused_workers)

    def full(self) -> bool:
        return self.capacity() == 0

    def has_worker(self, worker: bytes) -> bool:
        if worker in self._used_workers:
            return True

        if worker in self._unused_workers:
            return True

        return False

    def get_one_unused_worker(self) -> Optional[bytes]:
        if self.capacity():
            return next(iter(self._unused_workers))

        return None

    def get_unused_workers(self) -> Set[bytes]:
        return self._unused_workers

    def get_used_workers(self) -> Dict[bytes, bytes]:
        return self._used_workers



import logging
import multiprocessing
from multiprocessing.synchronize import Event as EventClass
import signal
import time
from typing import List

from scaled.protocol.python.serializer.mixins import Serializer
from scaled.utility.zmq_config import ZMQConfig
from scaled.utility.logging.utility import setup_logger
from scaled.worker.worker import Worker


class ClusterProcess(multiprocessing.get_context("spawn").Process):
    def __init__(
        self,
        stop_event: EventClass,
        address: ZMQConfig,
        n_workers: int,
        heartbeat_interval: int,
        event_loop: str,
        serializer: Serializer,
    ):
        multiprocessing.Process.__init__(self, name="WorkerMaster")
        assert isinstance(stop_event, EventClass)
        self._stop_event: multiprocessing.Event = stop_event

        self._address = address
        self._n_workers = n_workers
        self._heartbeat_interval = heartbeat_interval
        self._event_loop = event_loop
        self._serializer = serializer

        self._workers: List[Worker] = []

    def run(self):
        setup_logger()
        self.__register_signal()
        self.__start_workers_and_run_forever()

    def __shutdown(self, *args):
        assert args is not None
        logging.info(f"received signal, abort")
        self._stop_event.set()

    def __register_signal(self):
        signal.signal(signal.SIGINT, self.__shutdown)
        signal.signal(signal.SIGTERM, self.__shutdown)

    def __start_workers_and_run_forever(self):
        logging.info(
            f"{self.__get_prefix()} starting {self._n_workers} workers, heartbeat interval is "
            f"{self._heartbeat_interval} seconds"
        )

        self._workers = [
            Worker(
                address=self._address,
                stop_event=self._stop_event,
                heartbeat_interval_seconds=self._heartbeat_interval,
                event_loop=self._event_loop,
                serializer=self._serializer,
            )
            for _ in range(self._n_workers)
        ]

        if self._n_workers == 1:
            self._workers[0].run()
            return

        for worker in self._workers:
            worker.start()

        for i, worker in enumerate(self._workers):
            logging.info(f"Worker[{i}] started")

        while not self._stop_event.is_set():
            time.sleep(0.1)
            continue

        for i, worker in enumerate(self._workers):
            worker.join()
            logging.info(f"Worker[{i}] quited")

        logging.info(f"{self.__get_prefix()} shutdown")

    def __get_prefix(self):
        return f"{self.__class__.__name__}:"



import asyncio
import threading
import time

import psutil
import zmq.asyncio

from scaled.io.async_connector import AsyncConnector
from scaled.protocol.python.message import Heartbeat, MessageType, MessageVariant
from scaled.utility.zmq_config import ZMQConfig


class AsyncAgent:
    def __init__(
        self,
        stop_event: threading.Event,
        context: zmq.asyncio.Context,
        address: ZMQConfig,
        address_internal: ZMQConfig,
        heartbeat_interval_seconds: int,
    ):
        self._stop_event = stop_event

        self._connector_external = AsyncConnector(
            prefix="W",
            context=context,
            socket_type=zmq.DEALER,
            address=address,
            bind_or_connect="connect",
            callback=self.on_receive_external,
        )
        self._connector_internal = AsyncConnector(
            prefix="A",
            context=context,
            socket_type=zmq.PAIR,
            address=address_internal,
            bind_or_connect="bind",
            callback=self.on_receive_internal,
        )

        self._heartbeat = _WorkerHeartbeat(
            connector=self._connector_external, heartbeat_interval_seconds=heartbeat_interval_seconds
        )

    @property
    def identity(self):
        return self._connector_external.identity

    async def on_receive_external(self, message_type: MessageType, message: MessageVariant):
        await self._connector_internal.send(message_type, message)

    async def on_receive_internal(self, message_type: MessageType, message: MessageVariant):
        await self._connector_external.send(message_type, message)

    async def loop(self):
        while not self._stop_event.is_set():
            await asyncio.gather(
                self._heartbeat.routine(), self._connector_external.routine(), self._connector_internal.routine()
            )


class _WorkerHeartbeat:
    def __init__(self, connector: AsyncConnector, heartbeat_interval_seconds: int):
        self._heartbeat_interval_seconds = heartbeat_interval_seconds
        self._connector: AsyncConnector = connector
        self._process = psutil.Process()

        # minus heartbeat interval seconds to trigger very first heartbeat when launching
        self._start: float = time.time() - self._heartbeat_interval_seconds

    async def routine(self):
        if time.time() - self._start < self._heartbeat_interval_seconds:
            return

        await self._connector.send(
            MessageType.Heartbeat, Heartbeat(self._process.cpu_percent() / 100, self._process.memory_info().rss)
        )
        self._start = time.time()



import logging
import multiprocessing

from scaled.cluster.scheduler import SchedulerProcess
from scaled.protocol.python.serializer.default import DefaultSerializer
from scaled.protocol.python.serializer.mixins import Serializer
from scaled.scheduler.worker_manager.vanilla import AllocatorType
from scaled.utility.zmq_config import ZMQConfig
from scaled.cluster.cluster import ClusterProcess


class SchedulerClusterCombo:
    def __init__(
        self,
        address: ZMQConfig,
        n_workers: int,
        heartbeat_interval: int = 1,
        event_loop: str = "builtin",
        worker_timeout_seconds: int = 10,
        function_timeout_seconds: int = 60,
        allocator_type: AllocatorType = AllocatorType.Queued,
        serializer: Serializer = DefaultSerializer(),
    ):
        self._stop_event = multiprocessing.get_context("spawn").Event()
        self._cluster = ClusterProcess(
            stop_event=self._stop_event,
            address=address,
            n_workers=n_workers,
            heartbeat_interval=heartbeat_interval,
            event_loop=event_loop,
            serializer=serializer,
        )
        self._scheduler = SchedulerProcess(
            address=address,
            stop_event=self._stop_event,
            allocator_type=allocator_type,
            worker_timeout_seconds=worker_timeout_seconds,
            function_timeout_seconds=function_timeout_seconds,
        )

        self._cluster.start()
        self._scheduler.start()
        logging.info(f"{self.__get_prefix()} started")

    def __del__(self):
        self.shutdown()
        logging.info(f"{self.__get_prefix()} shutdown")

    def shutdown(self):
        self._stop_event.set()
        self._cluster.join()
        self._scheduler.join()

    def __get_prefix(self):
        return f"{self.__class__.__name__}:"



import unittest

from scaled.scheduler.worker_manager.allocators.one_to_one import WorkerCollection


class TestWorkerCollection(unittest.TestCase):
    def test_worker_collection(self):
        collection = WorkerCollection()
        self.assertEqual(collection.size(), 0)
        self.assertEqual(collection.capacity(), 0)

        collection[b"a"] = None
        self.assertEqual(collection.size(), 1)
        self.assertEqual(collection.capacity(), 1)

        collection[b"b"] = None
        self.assertEqual(collection.size(), 2)
        self.assertEqual(collection.capacity(), 2)

        collection[b"c"] = None
        self.assertEqual(collection.size(), 3)
        self.assertEqual(collection.capacity(), 3)

        collection[b"a"] = b"1"
        self.assertEqual(collection.size(), 3)
        self.assertEqual(collection.capacity(), 2)

        collection[b"b"] = b"2"
        self.assertEqual(collection.size(), 3)
        self.assertEqual(collection.capacity(), 1)

        collection[b"d"] = None
        self.assertEqual(collection.size(), 4)
        self.assertEqual(collection.capacity(), 2)

        collection[b"a"] = None
        self.assertEqual(collection.size(), 4)
        self.assertEqual(collection.capacity(), 3)

        collection.pop(b"d")
        self.assertEqual(collection.size(), 3)
        self.assertEqual(collection.capacity(), 2)

        collection.pop(b"b")
        self.assertEqual(collection.size(), 2)
        self.assertEqual(collection.capacity(), 2)

    def test_worker_collection_in(self):
        collection = WorkerCollection()
        collection[b"a"] = None
        self.assertTrue(collection.has_worker(b"a"))
        self.assertFalse(collection.has_worker(b"w"))



import hashlib
import threading
import uuid
from concurrent.futures import Future

from typing import Callable, Dict, Optional, Set, Tuple

import zmq

from scaled.utility.zmq_config import ZMQConfig
from scaled.io.sync_connector import SyncConnector
from scaled.protocol.python.serializer.mixins import FunctionSerializerType
from scaled.protocol.python.serializer.default import DefaultSerializer
from scaled.protocol.python.message import (
    FunctionRequest,
    FunctionRequestType,
    FunctionResponse,
    FunctionResponseType,
    MessageType,
    MessageVariant,
    MonitorRequest,
    MonitorResponse,
    Task,
    TaskEcho,
    TaskEchoStatus,
    TaskResult,
)


class Client:
    def __init__(self, config: ZMQConfig, serializer: FunctionSerializerType = DefaultSerializer()):
        self._stop_event = threading.Event()
        self._connector = SyncConnector(
            stop_event=self._stop_event,
            prefix="C",
            context=zmq.Context.instance(),
            socket_type=zmq.DEALER,
            bind_or_connect="connect",
            address=config,
            callback=self.__on_receive,
        )
        self._serializer = serializer

        self._function_to_function_id_cache: dict[Callable, bytes] = dict()
        self._ready_function_ids: Set[bytes] = set()
        self._task_id_to_futures: Dict[bytes, Future] = dict()

        self._statistics_future: Optional[Future] = None

    def __del__(self):
        self.disconnect()

    def submit(self, fn: Callable, *args) -> Future:
        function_id = self.__get_function_id(fn)

        task_id = uuid.uuid1().bytes
        task = Task(task_id, function_id, self._serializer.serialize_arguments(args))
        self._connector.send(MessageType.Task, task)

        future = Future()
        self._task_id_to_futures[task_id] = future
        return future

    def statistics(self):
        self._statistics_future = Future()
        self._connector.send(MessageType.MonitorRequest, MonitorRequest())
        return self._statistics_future.result()

    def disconnect(self):
        self._stop_event.set()

    def __on_receive(self, message_type: MessageType, message: MessageVariant):
        if message_type == MessageType.FunctionResponse:
            self.__on_function_response(message)
            return

        if message_type == MessageType.TaskEcho:
            self.__on_task_echo(message)
            return

        if message_type == MessageType.TaskResult:
            self.__on_task_result(message)
            return

        if message_type == MessageType.MonitorResponse:
            self.__on_statistics_response(message)
            return

        raise TypeError(f"Unknown {message_type=}")

    def __on_task_echo(self, task_echo: TaskEcho):
        assert task_echo.status in {TaskEchoStatus.SubmitOK, TaskEchoStatus.CancelOK}, (
            f"Unknown task status: " f"{task_echo=}"
        )

        if task_echo.task_id not in self._task_id_to_futures:
            return

        future = self._task_id_to_futures[task_echo.task_id]
        future.set_running_or_notify_cancel()

    def __on_function_response(self, response: FunctionResponse):
        assert response.status in {FunctionResponseType.OK, FunctionResponseType.Duplicated}
        self._ready_function_ids.add(response.function_id)

    def __on_task_result(self, result: TaskResult):
        if result.task_id not in self._task_id_to_futures:
            return

        future = self._task_id_to_futures.pop(result.task_id)
        future.set_result(self._serializer.deserialize_result(result.result))

    def __on_statistics_response(self, monitor_response: MonitorResponse):
        self._statistics_future.set_result(monitor_response.data)

    def __get_function_id(self, fn: Callable) -> bytes:
        if fn in self._function_to_function_id_cache:
            return self._function_to_function_id_cache[fn]

        function_id, function_bytes = self.__generate_function_id_bytes(fn)
        self._connector.send(
            MessageType.FunctionRequest, FunctionRequest(FunctionRequestType.Add, function_id, function_bytes)
        )

        while function_id not in self._ready_function_ids:
            continue

        self._function_to_function_id_cache[fn] = function_id
        return function_id

    def __generate_function_id_bytes(self, fn) -> Tuple[bytes, bytes]:
        function_bytes = self._serializer.serialize_function(fn)
        function_id = hashlib.md5(function_bytes).hexdigest().encode()
        return function_id, function_bytes



import abc
from typing import Any, Callable, Tuple, TypeVar


class Serializer(metaclass=abc.ABCMeta):
    @staticmethod
    @abc.abstractmethod
    def serialize_function(fn: Callable) -> bytes:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def deserialize_function(payload: bytes) -> Callable:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def serialize_arguments(args: Tuple[Any, ...]) -> bytes:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def deserialize_arguments(payload: bytes) -> Tuple[Any, ...]:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def serialize_result(result: Any) -> bytes:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def deserialize_result(payload: bytes) -> Any:
        raise NotImplementedError()


FunctionSerializerType = TypeVar("FunctionSerializerType", bound=Serializer)



from typing import Dict

from scaled.protocol.python.message import Task, TaskResult
from scaled.scheduler.mixins import ClientManager


class VanillaClientManager(ClientManager):
    def __init__(self):
        pass

    async def on_task_new(self, client: bytes, task: Task):
        pass

    async def on_task_cancel(self, client: bytes, task_id: bytes):
        pass

    async def on_task_done(self, result: TaskResult):
        pass

    async def routine(self):
        pass

    async def statistics(self) -> Dict:
        pass



import asyncio
import threading
import logging

from scaled.scheduler.client_manager.vanilla import VanillaClientManager
from scaled.scheduler.function_manager.vanilla import VanillaFunctionManager
from scaled.utility.zmq_config import ZMQConfig
from scaled.io.async_binder import AsyncBinder
from scaled.protocol.python.message import MessageType, MessageVariant, MonitorRequest, MonitorResponse
from scaled.scheduler.task_manager.vanilla import VanillaTaskManager
from scaled.scheduler.worker_manager.vanilla import AllocatorType, VanillaWorkerManager

PREFIX = "Scheduler:"


class Scheduler:
    def __init__(
        self,
        address: ZMQConfig,
        stop_event: threading.Event,
        allocator_type: AllocatorType,
        worker_timeout_seconds: int,
        function_timeout_seconds: int,
    ):
        self._address = address
        self._stop_event = stop_event

        self._binder = AsyncBinder(prefix="S", address=self._address)
        self._client_manager = VanillaClientManager()
        self._function_manager = VanillaFunctionManager(function_timeout_seconds=function_timeout_seconds)
        self._task_manager = VanillaTaskManager(stop_event=self._stop_event)
        self._worker_manager = VanillaWorkerManager(
            stop_event=self._stop_event, allocator_type=allocator_type, timeout_seconds=worker_timeout_seconds
        )

        self._binder.register(self.on_receive_message)
        self._function_manager.hook(self._binder)
        self._task_manager.hook(self._binder, self._function_manager, self._worker_manager)
        self._worker_manager.hook(self._binder, self._task_manager)

    async def on_receive_message(self, source: bytes, message_type: MessageType, message: MessageVariant):
        if message_type == MessageType.Heartbeat:
            await self._worker_manager.on_heartbeat(source, message)
            return

        if message_type ==  MessageType.MonitorRequest:
            await self.statistics(source, message)
            return

        if message_type == MessageType.Task:
            await self._task_manager.on_task_new(source, message)
            return

        if message_type ==  MessageType.TaskCancel:
            await self._task_manager.on_task_cancel(source, message.task_id)
            return

        if message_type ==  MessageType.TaskResult:
            await self._worker_manager.on_task_done(message)
            return

        if message_type ==  MessageType.FunctionRequest:
            await self._function_manager.on_function(source, message)
            return

        logging.error(f"{PREFIX} unknown {message_type} from {source=}: {message}")

    async def loop(self):
        logging.info("Scheduler started")
        while not self._stop_event.is_set():
            await asyncio.gather(
                self._binder.routine(),
                self._task_manager.routine(),
                self._function_manager.routine(),
                self._worker_manager.routine(),
            )
        logging.info("Scheduler quited")

    async def statistics(self, source: bytes, request: MonitorRequest):
        assert isinstance(request, MonitorRequest)
        stats = MonitorResponse(
            {
                "binder": await self._binder.statistics(),
                "task_manager": await self._task_manager.statistics(),
                "function_manager": await self._function_manager.statistics(),
                "worker_manager": await self._worker_manager.statistics(),
            }
        )
        await self._binder.send(source, MessageType.MonitorResponse, stats)



import logging
from collections import defaultdict
import time
from typing import Dict, Optional, Set

from scaled.io.async_binder import AsyncBinder
from scaled.protocol.python.message import (
    FunctionRequest,
    FunctionRequestType,
    FunctionResponse,
    FunctionResponseType,
    MessageType,
)
from scaled.scheduler.mixins import FunctionManager


class VanillaFunctionManager(FunctionManager):
    def __init__(self, function_timeout_seconds: int):
        self._function_id_to_function: Dict[bytes, bytes] = dict()
        self._function_id_to_alive_since: Dict[bytes, float] = dict()

        self._function_id_to_task_ids: Dict[bytes, Set[bytes]] = defaultdict(set)

        self._function_timeout_seconds = function_timeout_seconds

        self._binder: Optional[AsyncBinder] = None

    def hook(self, binder: AsyncBinder):
        self._binder = binder

    async def on_function(self, source: bytes, request: FunctionRequest):
        if request.type == FunctionRequestType.Check:
            await self.__on_function_check(source, request.function_id)
            return

        if request.type == FunctionRequestType.Add:
            await self.__on_function_add(source, request.function_id, request.content)
            return

        if request.type == FunctionRequestType.Request:
            await self.__on_function_request(source, request)
            return

        if request.type == FunctionRequestType.Delete:
            await self.__on_function_scheduler_delete(source, request.function_id)
            return

        logging.error(f"received unknown function request type {request=} from {source=}")

    async def __on_function_request(self, source: bytes, function_request: FunctionRequest):
        if function_request.function_id not in self._function_id_to_function:
            await self._binder.send(
                source,
                MessageType.FunctionResponse,
                FunctionResponse(FunctionResponseType.NotExists, function_request.function_id, b""),
            )
            return

        await self._binder.send(
            source,
            MessageType.FunctionResponse,
            FunctionResponse(
                FunctionResponseType.OK,
                function_request.function_id,
                self._function_id_to_function[function_request.function_id],
            ),
        )

    async def has_function(self, function_id: bytes) -> bool:
        return function_id in self._function_id_to_function

    async def on_task_use_function(self, task_id: bytes, function_id: bytes):
        self._function_id_to_alive_since[function_id] = time.time()
        self._function_id_to_task_ids[function_id].add(task_id)

    async def on_task_done_function(self, task_id: bytes, function_id: bytes):
        task_ids = self._function_id_to_task_ids[function_id]
        task_ids.remove(task_id)
        if not task_ids:
            self._function_id_to_task_ids.pop(function_id)

    async def __on_function_check(self, client: bytes, function_id: bytes):
        if function_id in self._function_id_to_function:
            await self.__send_function_response(client, function_id, FunctionResponseType.OK)
            return

        await self.__send_function_response(client, function_id, FunctionResponseType.NotExists)

    async def __on_function_add(self, client: bytes, function_id: bytes, function: bytes):
        self._function_id_to_alive_since[function_id] = time.time()
        self._function_id_to_function[function_id] = function

        if function_id in self._function_id_to_function:
            await self.__send_function_response(client, function_id, FunctionResponseType.Duplicated)

        await self.__send_function_response(client, function_id, FunctionResponseType.OK)

    async def __on_function_scheduler_delete(self, client: bytes, function_id: bytes):
        if function_id not in self._function_id_to_function:
            await self.__send_function_response(client, function_id, FunctionResponseType.NotExists)
            return

        if len(self._function_id_to_task_ids[function_id]) > 0:
            await self.__send_function_response(client, function_id, FunctionResponseType.StillHaveTask)
            return

        self._function_id_to_function.pop(function_id)
        self._function_id_to_alive_since.pop(function_id)

        self._function_id_to_task_ids.pop(function_id)

        await self.__send_function_response(client, function_id, FunctionResponseType.StillHaveTask)

    async def __send_function_response(self, client: bytes, function_id: bytes, response_type: FunctionResponseType):
        await self._binder.send(client, MessageType.FunctionResponse, FunctionResponse(response_type, function_id, b""))

    async def routine(self):
        now = time.time()
        dead_functions = [
            function_id
            for function_id, alive_since in self._function_id_to_alive_since.items()
            if now - alive_since > self._function_timeout_seconds and function_id not in self._function_id_to_task_ids
        ]

        for function_id in dead_functions:
            self._function_id_to_function.pop(function_id)
            self._function_id_to_alive_since.pop(function_id)

    async def statistics(self) -> Dict:
        return {"function_id_to_tasks": {k.decode(): len(v) for k, v in self._function_id_to_task_ids.items()}}



import json
import random

import unittest

from scaled.client.client import Client
from scaled.utility.zmq_config import ZMQConfig, ZMQType
from scaled.utility.logging.scoped_logger import ScopedLogger
from scaled.utility.logging.utility import setup_logger


def sleep_print(sec: int):
    # time.sleep(sec)
    return sec * 1


class TestClient(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()

    def test_client(self):
        # need server
        config = ZMQConfig(type=ZMQType.tcp, host="127.0.0.1", port=2345)
        client = Client(config=config)

        tasks = [random.randint(0, 100) for _ in range(10000)]
        with ScopedLogger(f"submit {len(tasks)} tasks"):
            futures = [client.submit(sleep_print, i) for i in tasks]

        with ScopedLogger(f"gather {len(futures)} results"):
            results = [future.result() for future in futures]

        self.assertEqual(results, tasks)
        client.disconnect()

    def test_monitor(self):
        config = ZMQConfig(type=ZMQType.tcp, host="127.0.0.1", port=2345)
        client = Client(config=config)
        print(json.dumps(client.statistics(), indent=4))
        client.disconnect()



import argparse
import multiprocessing
import os

from scaled.protocol.python.serializer.default import DefaultSerializer
from scaled.utility.event_loop import EventLoopType
from scaled.utility.zmq_config import ZMQConfig
from scaled.cluster.cluster import ClusterProcess

stop_event = multiprocessing.get_context("spawn").Event()


def get_args():
    parser = argparse.ArgumentParser(
        "standalone compute cluster", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--num-of-workers", "-n", type=int, default=os.cpu_count() - 1, help="number of workers in " "cluster"
    )
    parser.add_argument(
        "--heartbeat-interval", type=int, default=10, help="number of seconds to send heartbeat " "interval"
    )
    parser.add_argument(
        "--event-loop", default="builtin", choices=EventLoopType.allowed_types(), help="select event loop type"
    )
    parser.add_argument("address", type=ZMQConfig.from_string, help="scheduler address to connect to")
    return parser.parse_args()


def main():
    args = get_args()
    cluster = ClusterProcess(
        stop_event=stop_event,
        address=args.address,
        n_workers=args.num_of_workers,
        heartbeat_interval=args.heartbeat_interval,
        event_loop=args.event_loop,
        serializer=DefaultSerializer(),
    )
    cluster.run()



import threading
import logging
from collections import deque
from typing import Deque, Dict, Literal, Optional, Set, Tuple

from scaled.io.async_binder import AsyncBinder
from scaled.protocol.python.message import (
    MessageType,
    Task,
    TaskCancelEcho,
    TaskEcho,
    TaskEchoStatus,
    TaskResult,
    TaskStatus,
)
from scaled.scheduler.mixins import FunctionManager, TaskManager, WorkerManager


class VanillaTaskManager(TaskManager):
    def __init__(self, stop_event: threading.Event):
        self._stop_event = stop_event

        self._binder: Optional[AsyncBinder] = None
        self._function_manager: Optional[FunctionManager] = None
        self._worker_manager: Optional[WorkerManager] = None

        self._task_id_to_client: Dict[bytes, bytes] = dict()
        self._task_id_to_task: Dict[bytes, Task] = dict()

        self._running: Set[bytes] = set()
        self._canceling: Set[bytes] = set()
        self._unassigned: Deque[Tuple[bytes, Task]] = deque()

        self._failed_count: int = 0
        self._canceled_count: int = 0

    def hook(self, binder: AsyncBinder, function_manager: FunctionManager, worker_manager: WorkerManager):
        self._binder = binder
        self._function_manager = function_manager
        self._worker_manager = worker_manager

    async def routine(self):
        await self.__on_assign_tasks()

    async def statistics(self) -> Dict:
        return {
            "running": len(self._running),
            "canceling": len(self._canceling),
            "unassigned": len(self._unassigned),
            "failed": self._failed_count,
            "canceled": self._canceled_count,
        }

    async def on_task_new(self, client: bytes, task: Task):
        if not await self._function_manager.has_function(task.function_id):
            await self._binder.send(
                client, MessageType.TaskEcho, TaskEcho(task.task_id, TaskEchoStatus.FunctionNotExists)
            )
            return

        await self._binder.send(client, MessageType.TaskEcho, TaskEcho(task.task_id, TaskEchoStatus.SubmitOK))
        await self._function_manager.on_task_use_function(task.task_id, task.function_id)
        self._unassigned.append((client, task))

    async def on_task_reroute(self, task_id: bytes):
        assert task_id in self._task_id_to_client

        client = self._task_id_to_client.pop(task_id)
        task = self._task_id_to_task.pop(task_id)
        self._running.remove(task_id)

        self._unassigned.append((client, task))

    async def on_task_cancel(self, client: bytes, task_id: bytes):
        if task_id in self._canceling:
            logging.warning(f"already canceling: {task_id=}")
            await self._binder.send(
                client, MessageType.TaskCancelEcho, TaskCancelEcho(task_id, TaskEchoStatus.Duplicated)
            )
            return

        if task_id not in self._running:
            logging.warning(f"cannot cancel task is not running: {task_id=}")
            await self._binder.send(
                client, MessageType.TaskCancelEcho, TaskCancelEcho(task_id, TaskEchoStatus.Duplicated)
            )
            return

        self._running.remove(task_id)
        self._canceling.add(task_id)
        await self._binder.send(client, MessageType.TaskCancelEcho, TaskCancelEcho(task_id, TaskEchoStatus.CancelOK))
        await self._worker_manager.on_task_cancel(task_id)

    async def on_task_done(self, result: TaskResult):
        """job done can be success or failed"""
        if result.status == TaskStatus.Success:
            await self.__on_task_success(result)
            return

        if result.status == TaskStatus.Failed:
            await self.__on_task_failed(self._running, "fail", result)
            return

        if result.status == TaskStatus.Canceled:
            await self.__on_task_failed(self._canceling, "cancel", result)
            return

        raise ValueError(f"unknown TaskResult status: {result.status}")

    async def __on_assign_tasks(self):
        if not self._unassigned:
            return

        client, task = self._unassigned[0]
        if not await self._worker_manager.assign_task_to_worker(task):
            return

        self._unassigned.popleft()
        self._task_id_to_client[task.task_id] = client
        self._task_id_to_task[task.task_id] = task
        self._running.add(task.task_id)

    async def __on_task_success(self, result: TaskResult):
        if result.task_id not in self._running:
            return

        self._running.remove(result.task_id)
        task = self._task_id_to_task.pop(result.task_id)
        client = self._task_id_to_client.pop(result.task_id)

        await self._binder.send(client, MessageType.TaskResult, result)
        await self._function_manager.on_task_done_function(task.task_id, task.function_id)

    async def __on_task_failed(
        self, processing_set: Set[bytes], fail_type: Literal["cancel", "fail"], result: TaskResult
    ):
        if result.task_id not in processing_set:
            return

        processing_set.remove(result.task_id)
        task = self._task_id_to_task.pop(result.task_id)
        client = self._task_id_to_client.pop(result.task_id)

        if fail_type == "cancel":
            self._canceled_count += 1
        elif fail_type == "fail":
            self._failed_count += 1

        await self._binder.send(client, MessageType.TaskResult, result)
        await self._function_manager.on_task_done_function(task.task_id, task.function_id)



import asyncio
import multiprocessing
import unittest

from scaled.protocol.python.serializer.default import DefaultSerializer
from scaled.utility.zmq_config import ZMQConfig, ZMQType
from scaled.io.async_binder import AsyncBinder
from scaled.protocol.python.message import MessageType, MessageVariant
from scaled.worker.worker import Worker


class TestWorker(unittest.TestCase):
    def test_worker_heartbeat(self):
        async def callback(to: bytes, message_type: MessageType, message: MessageVariant):
            print(message)

        config = ZMQConfig(type=ZMQType.tcp, host="127.0.0.1", port=12346)

        stop_event = multiprocessing.get_context("spawn").Event()
        worker = Worker(
            stop_event=stop_event,
            address=config,
            heartbeat_interval_seconds=1,
            event_loop="builtin",
            serializer=DefaultSerializer(),
        )
        worker.start()

        driver = AsyncBinder(prefix="Backend", address=config)
        driver.register(callback)
        asyncio.run(driver.routine())



import logging
import os
import socket
from collections import defaultdict
from typing import Awaitable, Callable, Dict, List, Literal, Optional

import zmq.asyncio

from scaled.io.config import POLLING_TIME_MILLISECONDS
from scaled.utility.zmq_config import ZMQConfig
from scaled.protocol.python.message import MessageType, MessageVariant, PROTOCOL


class AsyncBinder:
    def __init__(self, prefix: str, address: ZMQConfig):
        self._address = address
        self._identity: bytes = f"{prefix}|{socket.gethostname()}|{os.getpid()}".encode()

        self._context = zmq.asyncio.Context.instance()
        self._socket = self._context.socket(zmq.ROUTER)
        self.__set_socket_options()
        self._socket.bind(self._address.to_address())
        logging.info(f"{self.__get_prefix()} bind to {address.to_address()}")

        self._callback: Optional[Callable[[bytes, MessageType, MessageVariant], Awaitable[None]]] = None

        self._statistics = {"received": defaultdict(lambda: 0), "sent": defaultdict(lambda: 0)}

    def register(self, callback: Callable[[bytes, MessageType, MessageVariant], Awaitable[None]]):
        self._callback = callback

    async def routine(self):
        count = await self._socket.poll(POLLING_TIME_MILLISECONDS)
        if not count:
            return

        for _ in range(count):
            frames = await self._socket.recv_multipart()
            if not self.__is_valid_message(frames):
                continue

            source, message_type_bytes, payload = frames[0], frames[1], frames[2:]
            message_type = MessageType(message_type_bytes)
            self.__count_one("received", message_type)
            message = PROTOCOL[message_type_bytes].deserialize(payload)
            await self._callback(source, message_type, message)

    async def statistics(self) -> Dict:
        return {
            "received": {k: v for k, v in self._statistics["received"].items()},
            "sent": {k: v for k, v in self._statistics["sent"].items()},
        }

    async def send(self, to: bytes, message_type: MessageType, message: MessageVariant):
        self.__count_one("sent", message_type)
        await self._socket.send_multipart([to, message_type.value, *message.serialize()])

    def __set_socket_options(self):
        self._socket.setsockopt(zmq.IDENTITY, self._identity)
        self._socket.setsockopt(zmq.SNDHWM, 0)
        self._socket.setsockopt(zmq.RCVHWM, 0)

    def __is_valid_message(self, frames: List[bytes]) -> bool:
        if len(frames) < 3:
            logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
            return False

        if frames[1] not in {member.value for member in MessageType}:
            logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
            return False

        return True

    def __count_one(self, count_type: Literal["sent", "received"], message_type: MessageType):
        self._statistics[count_type][message_type.name] += 1

    def __get_prefix(self):
        return f"{self.__class__.__name__}[{self._identity.decode()}]:"



import enum
from typing import Optional

import attrs
from attrs.validators import instance_of, optional


class ZMQType(enum.Enum):
    inproc = "inproc"
    tcp = "tcp"


@attrs.define
class ZMQConfig:
    type: ZMQType = attrs.field(validator=instance_of(ZMQType), converter=ZMQType)
    host: str = attrs.field(validator=instance_of(str))
    port: Optional[int] = attrs.field(validator=optional(instance_of(int)), default=None)

    def __attrs_post_init__(self):
        if self.type == ZMQType.inproc and self.port is not None:
            raise ValueError("inproc type should not have `port`")

    def to_address(self):
        if self.type == ZMQType.tcp:
            return f"tcp://{self.host}:{self.port}"

        if self.type == ZMQType.inproc:
            return f"inproc://{self.host}"

        raise TypeError(f"Unsupported ZMQ type: {self.type}")

    @staticmethod
    def from_string(string: str) -> "ZMQConfig":
        if "://" not in string:
            raise ValueError(f"valid ZMQ config should be like tcp://127.0.0.1:12345")

        socket_type, host_port = string.split("://", 1)
        if socket_type not in {member.value for member in ZMQType}:
            raise ValueError(f"supported ZMQ types are: {[member.value for member in ZMQType]}")

        host, port = host_port.split(":")
        return ZMQConfig(ZMQType(socket_type), host, int(port))



import enum
import logging


class EventLoopType(enum.Enum):
    builtin = enum.auto()
    uvloop = enum.auto()

    @staticmethod
    def allowed_types():
        return {m.name for m in EventLoopType}


def register_event_loop(event_loop_type: str):
    if event_loop_type not in EventLoopType.allowed_types():
        raise TypeError(f"allowed event loop types are: {EventLoopType.allowed_types()}")

    event_loop_type = EventLoopType[event_loop_type]
    if event_loop_type == EventLoopType.uvloop:
        try:
            import uvloop
        except ImportError:
            raise ImportError(f"please use pip install uvloop if try to use uvloop as event loop")
        uvloop.install()



    logging.info(f"use event loop: {event_loop_type.value}")



import argparse
import asyncio
import signal
import threading

from scaled.scheduler.main import Scheduler
from scaled.scheduler.worker_manager.vanilla import AllocatorType
from scaled.utility.event_loop import EventLoopType, register_event_loop
from scaled.utility.zmq_config import ZMQConfig
from scaled.utility.logging.utility import setup_logger

stop_event = threading.Event()


def get_args():
    parser = argparse.ArgumentParser("scaled scheduler", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "--worker-timeout-seconds", type=int, default=60, help="discard worker when timeout seconds " "reached"
    )
    parser.add_argument(
        "--function-timeout-seconds",
        type=int,
        default=60,
        help="discard function in scheduler when timeout seconds " "reached",
    )
    parser.add_argument(
        "--allocator-type",
        required=True,
        type=AllocatorType,
        choices={t for t in AllocatorType},
        help="specify allocator type",
    )
    parser.add_argument(
        "--event-loop", default="builtin", choices=EventLoopType.allowed_types(), help="select event loop type"
    )
    parser.add_argument("address", type=ZMQConfig.from_string, help="scheduler address to connect to")

    return parser.parse_args()


def main():
    args = get_args()
    setup_logger()

    __register_signal()

    scheduler = Scheduler(
        address=args.address,
        stop_event=stop_event,
        allocator_type=args.allocator_type,
        worker_timeout_seconds=args.worker_timeout_seconds,
        function_timeout_seconds=args.function_timeout_seconds,
    )
    register_event_loop(args.event_loop)
    asyncio.run(scheduler.loop())


def __register_signal():
    signal.signal(signal.SIGINT, __handle_signal)
    signal.signal(signal.SIGTERM, __handle_signal)


def __handle_signal(*args):
    assert args is not None
    stop_event.set()



import enum
import threading
import logging
import time
from typing import Dict, Optional

from scaled.io.async_binder import AsyncBinder
from scaled.scheduler.mixins import TaskManager, WorkerManager
from scaled.scheduler.worker_manager.allocators.one_to_one import OneToOneAllocator
from scaled.protocol.python.message import Heartbeat, MessageType, Task, TaskResult, TaskCancel
from scaled.scheduler.worker_manager.allocators.queued import QueuedAllocator

POLLING_TIME = 1


class AllocatorType(enum.Enum):
    OneToOne = "one_to_one"
    Queued = "queued"

    def __repr__(self):
        return self.value


class VanillaWorkerManager(WorkerManager):
    def __init__(self, stop_event: threading.Event, allocator_type: AllocatorType, timeout_seconds: int):
        self._stop_event = stop_event
        self._timeout_seconds = timeout_seconds

        self._binder: Optional[AsyncBinder] = None
        self._task_manager: Optional[TaskManager] = None

        self._worker_alive_since = {}

        if allocator_type == AllocatorType.OneToOne:
            self._allocator = OneToOneAllocator()
        elif allocator_type == AllocatorType.Queued:
            self._allocator = QueuedAllocator(1000)
        else:
            raise TypeError(f"received invalid allocator type: {allocator_type}")

    def hook(self, binder: AsyncBinder, task_manager: TaskManager):
        self._binder = binder
        self._task_manager = task_manager

    async def on_heartbeat(self, worker: bytes, info: Heartbeat):
        if self._allocator.add_worker(worker):
            logging.info(f"worker {worker} connected")

        self._worker_alive_since[worker] = time.time()

    async def assign_task_to_worker(self, task: Task) -> bool:
        worker = self._allocator.assign_task(task.task_id)
        if worker is None:
            return False

        # send to worker
        await self._binder.send(worker, MessageType.Task, task)
        return True

    async def on_task_cancel(self, task_id: bytes):
        worker = self._allocator.get_assigned_worker(task_id)
        if worker is None:
            logging.error(f"cannot find {task_id=} in task workers")
            return

        await self._binder.send(worker, MessageType.TaskCancel, TaskCancel(task_id))

    async def on_task_done(self, task_result: TaskResult):
        worker = self._allocator.remove_task(task_result.task_id)
        if worker is None:
            logging.error(
                f"received task_id={task_result.task_id} not known to any worker, might due to worker get disconnected"
            )
            return

        await self._task_manager.on_task_done(task_result)

    async def routine(self):
        await self.__clean_workers()

    async def statistics(self) -> Dict:
        return self._allocator.statistics()

    async def __clean_workers(self):
        now = time.time()
        dead_workers = [
            w for w, alive_since in self._worker_alive_since.items() if now - alive_since > self._timeout_seconds
        ]
        for dead_worker in dead_workers:
            logging.info(f"worker {dead_worker} disconnected")
            self._worker_alive_since.pop(dead_worker)

            task_ids = self._allocator.remove_worker(dead_worker)
            if not task_ids:
                continue

            logging.info(f"rerouting {len(task_ids)} tasks")
            for task_id in task_ids:
                await self._task_manager.on_task_reroute(task_id)
